import pandas as pd

from sqlalchemy import create_engine

from src.lib.formatters.sql import SQLFormatter
from src.tests.lib.writers.fixtures import *  # noqa: F403, F401
from src.tests.lib.writers.databases.fixtures import *  # noqa: F403, F401
from src.lib.writers.databases.postgresql import PostgresDirectDatabaseWriter


class TestPostgresDirectDatabaseWriter(object):
    def test_before_write(self, mocker, sql_specification_format):
        expected = "123"
        mock = mocker.patch('getpass._raw_input')
        mock.return_value = expected

        formatter = SQLFormatter(
            specification=sql_specification_format('mytable'))
        writer = PostgresDirectDatabaseWriter(formatter=formatter,
                                              specification={'options': {}})
        writer.before_write()

        mock.assert_called()
        assert writer.specification['options']['password'] == expected

    def test_writting_csv_with_records(self,
                                       mocker,
                                       pandas_dataframe_with_data,
                                       sql_specification_format,
                                       postgres_specification_direct):
        db_engine = create_engine('sqlite:///:memory:')

        mock = mocker.patch('getpass._raw_input')
        mock.return_value = "mytable"

        mock_ = mocker.patch.object(PostgresDirectDatabaseWriter, 'engine')
        mock_.return_value = db_engine

        formatter = SQLFormatter(
            specification=sql_specification_format('mytable'))
        writer = PostgresDirectDatabaseWriter(
            formatter=formatter,
            specification=postgres_specification_direct)
        writer.before_write()
        writer.write(dataframe=pandas_dataframe_with_data)

        dataframe_from_sql = pd.read_sql_table(
            table_name='mytable', con=db_engine)
        assert dataframe_from_sql.equals(pandas_dataframe_with_data)
        mock_.assert_called()

    def test_writting_dataframe_without_records(self,
                                                mocker,
                                                sql_specification_format,
                                                pandas_dataframe_without_data,
                                                postgres_specification_direct):
        mock_conn = mocker.patch('psycopg2.connect')
        mock_conn.cursor.return_value = lambda x: 0

        formatter = SQLFormatter(specification={})
        writer = PostgresDirectDatabaseWriter(
            formatter=formatter,
            specification=postgres_specification_direct)
        writer.write(dataframe=pandas_dataframe_without_data)

        mock_conn.assert_not_called()



import os


def get_version(package):
    if 'TRAVIS_TAG' in os.environ:
        return os.getenv("TRAVIS_TAG", 'v0.0')
    else:
        try:
            from importlib.metadata import version
            return version(package)
        except ModuleNotFoundError:
            import pkg_resources
            return pkg_resources.get_distribution(package).version



general_rules = {
    "datasets.{*}.size": {'type': int, 'none': False},
    "datasets.{*}.locale": {'type': str, 'none': False},
    "datasets.{*}.fields.[*].type": {'type': str, 'none': False},
    "datasets.{*}.fields.[*].name": {'type': str, 'none': False},
    "datasets.{*}.fields.[*].generator": {'type': dict, 'none': False},
    "datasets.{*}.format.type": {'type': str, 'none': False},
    "datasets.{*}.serializers.to.[*].type": {'type': str, 'none': False}
}

datetime_rules = {
    "format.options": {'type': dict, 'none': False},
    "format.options.date_format": {
        'type': str,
        'none': False,
        'values': ['iso', 'epoch']
    }
}



from s3fs.core import S3FileSystem
from pandas import DataFrame


class S3RemoteWriter(object):
    key = 's3'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification

    def write(self, dataframe: DataFrame) -> None:
        s3 = S3FileSystem(anon=False)
        options = self.specification['options']
        key = f's3://{options["bucket"]}/{options["key"]}'
        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=s3.open(key, mode='w'))

        return key

    @staticmethod
    def rules():
        return {'required': {'options.bucket': {'none': False, 'type': str},
                             'options.key': {'none': False, 'type': str}},
                'optional': {}}



import boto3

from moto import mock_s3

from src.lib.formatters.csv import CsvFormatter
from src.lib.writers.remotes.s3 import S3RemoteWriter
from src.tests.lib.writers.remotes.fixtures import *  # noqa: F403, F401
import pandas as pd


class TestS3Writer(object):
    """Unit-test for S3Writer class"""

    def test_writting_csv_with_records(self,
                                       pandas_dataframe_with_data,
                                       specification_s3):
        with mock_s3():
            bucket = specification_s3['options']['bucket']
            key = specification_s3['options']['key']
            conn = boto3.resource('s3', region_name='us-east-1')
            conn.create_bucket(Bucket=bucket)

            formatter = CsvFormatter(specification={})
            writer = S3RemoteWriter(formatter=formatter,
                                    specification=specification_s3)
            writer.write(dataframe=pandas_dataframe_with_data)

            body = conn.Object(bucket, key).get()['Body']

            csv_reader = pd.read_csv(body).to_dict(orient="list")
            expected = pandas_dataframe_with_data.to_dict(orient="list")

            assert csv_reader == expected

    def test_writting_csv_without_records(self,
                                          pandas_dataframe_without_data,
                                          specification_s3):
        with mock_s3():
            bucket = specification_s3['options']['bucket']
            key = specification_s3['options']['key']
            conn = boto3.resource('s3', region_name='us-east-1')
            conn.create_bucket(Bucket=bucket)

            formatter = CsvFormatter(specification={})
            writer = S3RemoteWriter(formatter=formatter,
                                    specification=specification_s3)
            writer.write(dataframe=pandas_dataframe_without_data)

            object_ = conn.Object(bucket, key).get()

            assert object_['ContentLength'] == 0



from pandas import DataFrame


class FileWriter(object):
    """Class that receive pandas dataframe
    and write it down in Json format
    """
    key = 'file'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification

    def write(self, dataframe: DataFrame) -> None:
        """Write all dataframe on a .json file.

        Keyword arguments:
         - dataframe - pandas.Dataframe: dataframe containing the records
        """
        file_path = self.specification['uri']
        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=file_path)
        return file_path

    @staticmethod
    def rules():
        return {'required': {'uri': {'none': False, 'type': str}},
                'optional': {}}



from azure.storage.blob import BlobServiceClient
from os import getenv
from pandas import DataFrame
from io import StringIO


class AzureBSRemoteWriter(object):
    key = 'azure-bs'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification
        self.azure_bs = self.__azure_storage_client()

    def __azure_storage_client(self):
        connection_string = self.get_credentials()
        if not connection_string:
            raise Exception("AZURE_STORAGE_CONNECTION_STRING Not found. "
                            "Please set your environment variable "
                            "with your connection string")
        return BlobServiceClient\
            .from_connection_string(conn_str=connection_string)

    def get_credentials(self):
        return getenv("AZURE_STORAGE_CONNECTION_STRING", "")

    def write(self, dataframe: DataFrame) -> None:
        buffer = self.open_buffer()
        self.formatter.format(dataframe, buffer)

        options = self.specification['options']
        container = options['container']
        blob = options['blob']

        self.upload_blob(container=container, blob=blob, data=buffer)

        return f'{options["container"]}/{options["blob"]}'

    def open_buffer(self):
        buffer = StringIO()
        return buffer

    def upload_blob(self, container, blob, data: StringIO):
        self.azure_bs.get_blob_client(container=container,
                                      blob=blob)\
            .upload_blob(data.getvalue(), overwrite=True)

    @staticmethod
    def rules():
        return {'required': {'options.container': {'none': False, 'type': str},
                             'options.blob': {'none': False, 'type': str}},
                'optional': {}}



from subprocess import Popen, PIPE


class Shell(object):
    def __init__(self):
        pass

    def execute(self, command):
        returncode, output, error = self.execute_command(command)
        if returncode > 0 and error:
            raise ShellError(error.decode())

        return output.decode()

    def execute_command(self, command):
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        output, error = process.communicate()
        return process.returncode, output, error


class ShellError(Exception):
    pass



import click
from src.generators.handler import GeneratorsHandler
from src.generators.dryrunhandler import DryRunHandler
from src.cli.ascii_art import ASCII_ART
from src.generators.datatypes import Metadata
from tabulate import tabulate
from locale import getdefaultlocale
from src import utils
from src.cli import get_version

VERSION = get_version('ggenerator')
default_locale = getdefaultlocale()[0]


@click.group()
@click.version_option(version=VERSION)
def execute():
    click.echo(ASCII_ART)
    click.echo(f"version: {VERSION}")
    click.echo("-------------------")


@execute.command()
@click.option('-s', '--spec', 'spec_path',
              type=click.Path(exists=True,
                              resolve_path=True,
                              file_okay=True,
                              dir_okay=False),
              required=True)
@click.option("--dryrun", "dryrun_flag",
              is_flag=True, default=False,
              required=False,
              help="This generate a sample of 10 records by the spec")
def generate(spec_path, dryrun_flag):
    if dryrun_flag:
        generate_dryrun(spec_path)
    else:
        generate_datasets(spec_path)


def generate_datasets(spec_path):
    try:
        generator = \
            GeneratorsHandler(arguments={'config_file': spec_path})
        for dset_name, dset_format, dset_path in generator.generate():
            click.echo("| Finished!\n"
                       f"| Dataset name: {dset_name}\n"
                       f"| Dataset format: {dset_format}\n"
                       f"| Dataset: {dset_path}\n")
    except Exception as err:
        click.echo(f"Error: {err}")


def generate_dryrun(spec_path):
    try:
        dryrun = DryRunHandler(arguments={'config_file': spec_path})
        dryrun.generate()
    except Exception as err:
        click.echo(f"Error: {err}")


@execute.command()
@click.option("--locale", "locale",
              type=click.STRING,
              required=False,
              default=default_locale)
def list_generators(locale):
    metadata = Metadata(locale=locale)
    infos = metadata.info()
    print_tabulate(dataframe=infos, headers=infos.columns,
                   tablefmt="fancy_grid")


def print_tabulate(dataframe, headers, tablefmt):
    click.echo(tabulate(dataframe, headers=headers, tablefmt=tablefmt))


@execute.command("generate-sample")
@click.option("--type", "typename", type=click.STRING, required=True)
@click.option("--locale", "locale", type=click.STRING, required=False,
              default=default_locale)
@click.option("--params", "params", multiple=True, required=False)
def sample(typename, locale, params):
    params = utils.args_to_kwargs(params=params, separator='=')
    metadata = Metadata(locale=locale)

    try:
        sample = metadata.sample(typename, **params)
        click.echo(f"Sample: {sample}")
    except Exception as err:
        click.echo(f"Error: {err}")


def get_uri(dataset_name, output_type):
    return click.prompt(f"Please, enter a valid URI of destination "
                        f"for the dataset: {dataset_name} "
                        f"and destination: {output_type}", type=str)



import inspect
from src.generators.datatypes.sequence_timestamp import TimestampSequenceType
from src.generators.datatypes.sequence import SequenceType
from src.generators.datatypes.fakerproxy import FakerProxy
from pandas import DataFrame, Series


class Metadata(object):
    def __init__(self, locale):
        self.locale = locale
        self.generators = (TimestampSequenceType, SequenceType)
        self.faker = FakerProxy(locale=self.locale)

    def info(self) -> DataFrame:
        return self.__dataframe_info()

    def sample(self, datatype, *args, **kwargs):
        try:
            return self.__sample(datatype, *args, **kwargs)
        except Exception as err:
            raise Exception(err)

    def __sample(self, datatype, *args, **kwargs):
        generators_map = self.get_generators_map()
        try:
            sample = generators_map[datatype]['type'](**kwargs).generate()
            return sample
        except KeyError:
            raise TypeError("Type not found\n"
                            "Check the specification to use a valid one\n"
                            "List all the types using "
                            "the command list-generators")
        except TypeError:
            raise TypeError(
                "Invalid parameter, "
                "please check the spec to correct name and value type")

        except Exception:
            raise ValueError(
                "Invalid parameter, "
                "please check the spec to correct name and value type")

    def __dataframe_info(self):
        df = DataFrame(columns=['type', 'namespace', 'parameters'])
        for key, value in self.get_generators_map().items():
            info = Series(
                {
                    'namespace': value['namespace'],
                    'type': key,
                    'parameters': value['generator']['arguments']
                })
            df = df.append(info, ignore_index=True)
        return df

    def get_generators_map(self) -> dict:
        generators_map = {
            generator.key: {
                'type': generator,
                'namespace': generator.namespace,
                'generator': {
                    'optional': generator.optional_arguments,
                    'arguments': self.__get_args_info(generator)
                }
            }
            for generator in self.generators}

        faker_types = {
            dtype.__name__: {
                'type': self.faker.__getattribute__(dtype.__name__),
                'namespace': self.faker.__getattribute__(dtype.__name__)
                                       .namespace,
                'generator': {
                    'optional': self.faker.__getattribute__(dtype.__name__)
                                          .optional_arguments,
                    'arguments': self.__get_args_info(
                        self.faker.__getattribute__(dtype.__name__)
                                  .generator_type
                    )
                }
            }
            for dtype in self.faker.list_of_generators
        }

        generators_map.update(faker_types)
        return generators_map

    def __get_args_info(self, func):
        params = self.__get_parameters(func)
        infos = []
        for arg_name, arg_info in params.items():
            info = f"{arg_name}"
            info += f" | default: {arg_info.default}" \
                if arg_info.default is not inspect._empty\
                else ''

            info += f" | type: {arg_info.annotation.__name__ }"\
                if arg_info.annotation is not inspect._empty\
                else ''

            infos.append(info)

        return '\n'.join(infos)

    def __get_parameters(self, func):
        return inspect.signature(func).parameters



from pytest import fixture


@fixture
def generate_files():
    yield "dset_name", "dset_format", "dset_path"


@fixture
def generators_map():
    return {
        "integer": {
            'type': "integer",
            'namespace': "basic_type",
            'generator': {
                    'optional': True,
                    'arguments': "start_at | end_at"
            }
        },
        "string": {
            'type': "string",
            'namespace': "basic_type",
            'generator': {
                    'optional': True,
                    'arguments': "length"
            }
        }
    }



from csv import reader
from os import remove
from os.path import exists, isfile

from src.lib.formatters.csv import CsvFormatter
from src.tests.lib.writers.fixtures import *  # noqa: F403, F401


class TestCsvWriter:
    """Unit-test of CsvWriter class"""

    def test_writting_dataframe_with_records(self,
                                             pandas_dataframe_with_data,
                                             specification):
        test_file_path = specification['uri']
        csv_writer = CsvFormatter(specification={"options": {"index": False}})
        csv_writer.format(dataframe=pandas_dataframe_with_data,
                          path_or_buffer=test_file_path)
        assert exists(test_file_path) is True
        assert isfile(test_file_path) is True

        with open(test_file_path) as csv_file:
            csv_reader = reader(csv_file, delimiter=',')
            csv_reader = [x for x in csv_reader]
            headers = csv_reader[0]
            rows = csv_reader[1:len(csv_reader)]
            csv_reader = {headers[index]: [row[index]
                                           for row in rows]
                          for index in range(0, len(headers))}
            expected = pandas_dataframe_with_data.to_dict(orient="list")

            assert csv_reader == expected

        remove(test_file_path)
        assert exists(test_file_path) is False

    def test_writting_dataframe_without_records(self,
                                                pandas_dataframe_without_data,
                                                specification):
        test_file_path = specification['uri']
        csv_writer = CsvFormatter(specification={"options": {"index": False}})
        csv_writer.format(dataframe=pandas_dataframe_without_data,
                          path_or_buffer=test_file_path)
        assert exists(test_file_path) is False
        assert isfile(test_file_path) is False



from pandas import DataFrame
from sqlalchemy.engine import Engine
from src.lib.sql import Sql


class SQLFormatter(object):
    """Class that receive pandas dataframe
    and write it down in .sql format
    """
    key = 'sql'
    modes = {'append': 'append',
             'truncate': 'append',
             'replace': 'replace'}

    def __init__(self, specification):
        self.default = {
            'mode': 'append',
            'batch_size': 50,
            'index': False
        }
        self.specification = specification

    @staticmethod
    def rules():
        def replace_rule(options):
            mode = options.get("mode")
            if mode == "replace":
                if "schema" not in options.keys():
                    raise ValueError(
                        "Schema field is required for replace mode,\n"
                        " please insert it")
                schema = options.get("schema")
                for key in schema.keys():
                    field = schema.get(key)
                    if "sqltype" not in field.keys():
                        raise ValueError(
                            "The replace mode needs "
                            "'sqltype' in Schema fields")

        return {
            'required': {
                'options.table_name': {'none': False, 'type': str}
            },
            'optional': {
                'options.batch_size': {'none': False, 'type': int},
                'options.index': {'none': False, 'type': bool},
                'options.index_label': {'none': False, 'type': str},
                'options.mode': {'none': False,
                                 'type': str,
                                 'values': ["append", "replace", "truncate"]},
                'options.schema': {'none': False,
                                   'type': dict,
                                   'custom': []},
                'options': {'none': False,
                            'type': dict,
                            'custom': [replace_rule]}
            }
        }

    def __to_db(self,
                dataframe: DataFrame,
                conn: Engine,
                params,
                **kwargs) -> str:
        table_name = params.get("table_name")
        batch_size = params.get("batch_size")
        mode = params.get("mode", 'append')
        index_flag = params.get("index")
        index_label = params.get("index_label")

        try:
            if mode == 'truncate' and conn.has_table(table_name=table_name):
                conn.execution_options(autoCommit=True)\
                    .execute(f"""TRUNCATE TABLE {table_name}""")
            dataframe.to_sql(con=conn,
                             name=table_name,
                             if_exists=self.modes.get(mode),
                             chunksize=batch_size,
                             index=index_flag,
                             index_label=index_label,
                             **kwargs)
        except Exception as err:
            msg = ("Error: Check your credentials (username,"
                   " password, host, port, database)\n")
            raise ValueError(msg, err)

    def __to_sql(self,
                 dataframe: DataFrame,
                 path_or_buffer,
                 params,
                 **kwargs) -> str:
        sql = Sql()
        mode = params.get("mode")

        if params.get('index'):
            dataframe.index.name = params.get('index_label')
            dataframe = dataframe.reset_index(level=0)

        if mode == "append":
            data = sql.create_append_statement(dataframe, params)
        elif mode == "replace":
            data = sql.replace_statement(dataframe, params)
        elif mode == "truncate":
            data = sql.truncate_statement(dataframe, params)
        else:
            raise ValueError(f"Mode `{mode}` is invalid, expected:"
                             " 'append', 'truncate' or 'replace'")

        if isinstance(path_or_buffer, str):
            with open(path_or_buffer, 'w') as f:
                f.write(data)
        elif path_or_buffer is None:
            return data
        else:
            path_or_buffer.write(data)

    def format(self,
               dataframe: DataFrame,
               path_or_buffer,
               method=None,
               **kwargs) -> str:
        """Format dataframe to sql script.

        Parameters:
         - dataframe - pandas.DataFrame: dataframe containing the records.
         - path_or_buffer: path-like string or a File handler
        """
        parameters = self.default
        options = self.specification.get('options', {})
        parameters.update(options)
        mode = parameters.get('mode')

        if mode == 'replace':
            schema_columns = set(parameters.get('schema', dict()).keys())
            columns = set(dataframe.columns)

        if mode == 'replace' and len(schema_columns - columns) > 0:
            raise ValueError(f"{schema_columns - columns} "
                             "Column(s) not declared on Fields properties!\n"
                             "Schema and Fields must have the same columns")

        if mode == 'replace' and len(columns - schema_columns) > 0:
            raise ValueError(f"{columns - schema_columns} "
                             "Column(s) not declared on Schema properties!\n"
                             "Schema and Fields must have the same columns")

        if dataframe.shape[0] > 0:
            if method == 'direct':
                return self.__to_db(dataframe=dataframe,
                                    conn=path_or_buffer,
                                    params=parameters,
                                    **kwargs)
            return self.__to_sql(dataframe=dataframe,
                                 path_or_buffer=path_or_buffer,
                                 params=parameters,
                                 **kwargs)



from src.lib.writers import writers, database_writers


class SerializersConfiguration(object):

    @staticmethod
    def get_database_type(serializer):
        options = serializer.get('options', {})
        engine = options.get('engine')
        method = options.get('method')
        return f"{engine}-{method}"

    @staticmethod
    def get_rules(serializer):
        type_ = serializer.get('type')

        if type_ not in writers:
            type_ = SerializersConfiguration.get_database_type(serializer)
            if type_ not in database_writers:
                raise ValueError(f"Serializer type `{type_}` not found")

        type_class = writers.get(type_)

        return type_class.rules()



from src.lib.formatters.sql import SQLFormatter
from src.lib.postgresql.psql import PostgresSqlPsql
from src.lib.writers.databases.postgresql import PostgreSqlClientDatabaseWriter
from src.tests.lib.writers.fixtures import *  # noqa: F403, F401
from src.tests.lib.writers.databases.fixtures import *  # noqa: F403, F401


class TestPostgreSqlClientDatabaseWriter(object):
    def test_before_write(self, mocker, sql_specification_format):
        expected = "123"
        mock = mocker.patch('getpass._raw_input')
        mock.return_value = expected

        formatter = SQLFormatter(
            specification=sql_specification_format('mytable'))
        writer = PostgreSqlClientDatabaseWriter(
            formatter=formatter,
            specification={'options': {}}
        )
        writer.before_write()

        mock.assert_called()
        assert writer.specification['options']['password'] == expected

    def test_writting_csv_with_records(self,
                                       mocker,
                                       pandas_dataframe_with_data,
                                       sql_specification_format,
                                       postgres_specification_cli):
        mock_conn = mocker.patch.object(PostgresSqlPsql, 'execute_query')
        mock_conn.return_value = lambda x: None

        formatter = SQLFormatter(
            specification=sql_specification_format('mytable'))
        writer = PostgreSqlClientDatabaseWriter(
            formatter=formatter,
            specification=postgres_specification_cli
        )
        writer.write(dataframe=pandas_dataframe_with_data)

        mock_conn.assert_called()

    def test_writting_dataframe_without_records(self,
                                                mocker,
                                                pandas_dataframe_without_data,
                                                postgres_specification_cli):
        mock_conn = mocker.patch.object(PostgresSqlPsql, 'execute_query')
        mock_conn.return_value = lambda x: None

        formatter = SQLFormatter(specification={})
        writer = PostgreSqlClientDatabaseWriter(
            formatter=formatter,
            specification=postgres_specification_cli
        )
        writer.write(dataframe=pandas_dataframe_without_data)

        mock_conn.assert_not_called()



from pandas import DataFrame

from src.lib.formatters import formatters
from src.lib.writers import writers
from src.generators.basehandler import BaseHandler
from src.lib.config.serializers import SerializersConfiguration


class GeneratorsHandler(object):

    def __init__(self, arguments: dict):
        self.file_path = arguments['config_file']
        self.base = BaseHandler()
        self.specification = self.valid_specification_dataset()
        self.writers = writers

    def valid_specification_dataset(self):
        valid = self.base.valid_specification(self.file_path)
        return valid

    def generate(self):
        datasets = self.specification.get('datasets')
        for key in datasets.keys():
            dataset = datasets[key]
            dataset_format = dataset['format']
            dataframe = self.generate_dataframe(dataset)
            for destination in dataset['serializers']['to']:
                file_path = self.write_dataframe(dataframe,
                                                 destination,
                                                 dataset_format)
                file_format = dataset_format['type']
                yield key, file_format, file_path

    def generate_dataframe(self, specification: dict) -> DataFrame:
        size = specification['size']
        dataframe = self.base.generate_dataframe(specification, size)
        return dataframe

    def __before_format(self, formatter):
        '''
           Supposed to checks if a method called 'before_format'
        exists and if it does, call it.
        '''
        pass

    def __before_write(self, writer):
        '''
           Checks if a method called 'before_write'
        exists and if it does, call it.
        '''
        method = 'before_write'

        if hasattr(writer, method) and \
           callable(writer.before_write):
            writer.before_write()

    def write_dataframe(self,
                        dataframe: DataFrame,
                        destination: dict,
                        format_: dict) -> str:
        file_format = format_['type']
        formatter_class = formatters[file_format]
        formatter = formatter_class(specification=format_)
        self.__before_format(formatter)

        destination_type = destination['type']

        if destination_type not in writers:
            destination_type = \
                SerializersConfiguration.get_database_type(destination)

        writer_class = writers[destination_type]
        writer = writer_class(formatter=formatter, specification=destination)
        self.__before_write(writer)
        return writer.write(dataframe=dataframe)



from src.lib.formatters import formatters


class FormattersConfiguration(object):

    @staticmethod
    def get_rules(format):
        format_ = format
        type_ = format_.get('type')

        if type_ not in formatters:
            raise ValueError(f"Format type `{type_}` not found")

        type_class = formatters.get(type_)

        return type_class.rules()



from pytest import fixture


def pandas_dataframe(data):
    from pandas import DataFrame

    dataframe = DataFrame(data=data)

    return dataframe


@fixture
def pandas_dataframe_with_data():
    data = [{"Column": "Value_1", "Column2": "Value2_1"},
            {"Column": "Value_2", "Column2": "Value2_2"},
            {"Column": "Value_3", "Column2": "Value2_3"}]
    return pandas_dataframe(data=data)


@fixture
def pandas_dataframe_without_data():
    return pandas_dataframe(data=[])


@fixture
def specification_s3():
    from uuid import uuid4

    file_name = uuid4().hex

    return {
        'type': 's3',
        'options': {
            'bucket': 'mybucket',
            'key': file_name
        }
    }


@fixture
def specification_gcs():
    from uuid import uuid4

    file_name = uuid4().hex

    return {
        'type': 'gcs',
        'options': {
            'bucket': 'mybucket',
            'key': file_name
        }
    }


@fixture
def specification_azure_bs():
    from uuid import uuid4

    file_name = uuid4().hex

    return {
        'type': 'azure-bs',
        'options': {
            'container': 'mycontainer',
            'blob': file_name
        }
    }


@fixture
def specification_with_url():
    return {
        'type': 's3-url',
        'options': {
            "path": "file/path/example.json"
        }
    }


@fixture
def specification_url_gcs():
    return {
        'type': 'gcs-url',
        'uri': 'https://signed-url.example.com'
    }


@fixture
def specification_without_url():
    return {'type': 's3-url'}


@fixture
def gcs_without_url():
    return {'type': 's3-url'}


@fixture
def signed_post_file():
    return '{"url": "s3.url.com", "fields": {}}'


@fixture
def s3_without_options():
    return {
        'type': 's3'
    }


@fixture
def s3_without_bucket():
    return {
        "type": "s3",
        "options": {
            "key": "/path/to/file.json"
        }
    }



from pytest import fixture


def pandas_dataframe(data):
    from pandas import DataFrame

    dataframe = DataFrame(data=data)

    return dataframe


@fixture
def pandas_dataframe_with_data():
    data = [{"Column1": "Value_1", "Column2": "Value_12"},
            {"Column1": "Value_2", "Column2": "Value_12"},
            {"Column1": "Value_3", "Column2": "Value_12"},
            {"Column1": "Value_4", "Column2": "Value_12"},
            {"Column1": "Value_5", "Column2": "Value_12"},
            {"Column1": "Value_6", "Column2": "Value_12"}]
    return pandas_dataframe(data=data)


@fixture
def pandas_dataframe_without_data():
    return pandas_dataframe(data=[])


@fixture
def sql_specification_format():
    def _spec(table_name, index=False, index_label=None, mode='append'):
        return {
            'options': {
                'table_name': table_name,
                'batch_size': 2,
                'schema': {
                    'Column1': {
                        'quoted': True,
                        'sqltype': "VARCHAR(50)"
                    },
                    'Column2': {
                        'quoted': True,
                        'sqltype': "VARCHAR(50)"
                    },
                },
                'mode': mode,
                'index': index,
                'index_label': index_label
            }
        }
    return _spec


@fixture
def fixture_spec_default():
    specification = {
        "datasets": {
            "sample": {
                "size": 10,
                "locale": "pt_BR",
                "fields": [
                    {
                        "type": "integer:sequence",
                        "name": "id",
                        "generator": {
                            "start_at": 1
                        }
                    },
                    {
                        "type": "name",
                        "name": "nome",
                        "generator": {}
                    },
                    {
                        "type": "pyint",
                        "name": "idade",
                        "generator": {
                            "max_value": 120
                        }
                    },
                    {
                        "type": "pyfloat",
                        "name": "peso",
                        "generator": {
                            "positive": 'false',
                            "min_value": 0,
                            "max_value": 250
                        }
                    },
                    {
                        "type": "job",
                        "name": "trabalho",
                        "generator": {}
                    },
                    {
                        "type": "future_datetime",
                        "name": "data",
                        "generator": {}
                    }
                ],
                "format": {
                    "type": "sql",
                    "options": {
                        "index": 'false',
                        "table_name": "My_table",
                        "mode": "append"
                    }
                },
                "serializers": {
                    "to": [
                        {
                            "type": "file",
                            "uri": "/home/tadeu/Desktop/dataset.sql"
                        }
                    ]
                }
            }
        }
    }

    return specification


@fixture
def fixture_spec_replace():
    return {
        "options": {
            "table_name": "My_table",
            "mode": "replace",
            "schema": {
                "name": {
                    "sqltype": "VARCHAR(50)",
                    "quoted": True
                },
                "age": {
                    "sqltype": "INTEGER NOT NULL",
                    "quoted": False
                },
                "weight": {
                    "sqltype": "INTEGER NOT NULL",
                    "quoted": False
                },
                "job": {
                    "sqltype": "VARCHAR(50)",
                    "quoted": True
                },
                "datetime": {
                    "sqltype": "DATETIME",
                    "quoted": False
                }
            }
        }
    }


@fixture
def fixture_spec_replace_without_schema():
    return {
        "options": {
            "table_name": "My_table",
            "mode": "replace"
        }
    }


@fixture
def fixture_spec_append2():
    return {
        "options": {
            "table_name": "My_table",
            "mode": "append"
            }
        }


@fixture
def fixture_spec_replace_schema_changed():
    true = True
    specification = {"options": {
        "table_name": "My_table",
        "mode": "replace",
        "index": true,
        "index_label": "my_index_column",
        "schema": {
            "name": {"sqltype": "VARCHAR(50)", "quoted":
                     true}
        }

    }
    }

    return specification


@fixture
def replace_dataframe():
    data = [
        {
            "name": "Sophie Silveira",
            "age": 65,
            "weight": 194.16350529,
            "job": "Fisioterapeuta",
            "datetime": 1587241737000
        },
        {
            "name": "Davi Lucca Duarte",
            "age": 105,
            "weight": 19.41,
            "job": "Cabo",
            "datetime": 1588771828000
        },
        {
            "name": "Luna da Rosa",
            "age": 56,
            "weight": 104.987,
            "job": "Faxineiro",
            "datetime": 1588995002000
        },
        {
            "name": "Vitor Melo",
            "age": 74,
            "weight": 170.79,
            "job": "Sociólogo",
            "datetime": 1587715890000
        },
        {
            "name": "Catarina Correia",
            "age": 6,
            "weight": 178.9275,
            "job": "Médico geneticista",
            "datetime": 1588948730000
        }
    ]
    return pandas_dataframe(data=data)


@fixture
def dataframe_with_datetime():
    from dateutil.parser import isoparse

    data = [{"Column": "Value_1", "at": "2020-10-01 00:00:00"},
            {"Column": "Value_2", "at": "2020-10-01 00:00:00"},
            {"Column": "Value_3", "at": "2020-10-01 00:00:00"}]
    df = pandas_dataframe(data=data)
    df['at'] = df['at'].apply(lambda x: isoparse(x))

    return df


@fixture
def invalid_spec_for_replace_schema_bigger_than_fields():
    false = False
    true = True

    return {
        "datasets": {
            "sample": {
                "size": 10,
                "locale": "pt_BR",
                "fields": [
                    {
                        "type": "integer:sequence",
                        "name": "id",
                        "generator": {
                            "start_at": 1
                        }
                    },
                    {
                        "type": "pyint",
                        "name": "age",
                        "generator": {
                            "max_value": 120
                        }
                    },
                    {
                        "type": "pyfloat",
                        "name": "weight",
                        "generator": {
                            "positive": false,
                            "min_value": 0,
                            "max_value": 250
                        }
                    },
                    {
                        "type": "future_datetime",
                        "name": "datetime",
                        "generator": {}
                    }
                ],
                "format": {
                    "type": "sql",
                    "options": {
                        "table_name": "My_table",
                        "mode": "replace",
                        "schema": {
                            "id": {
                                "sqltype": "INTEGER NOT NULL",
                                "quoted": false
                            },
                            "name": {
                                "sqltype": "VARCHAR(50)",
                                "quoted": true
                            },
                            "age": {
                                "sqltype": "INTEGER NOT NULL",
                                "quoted": false
                            },
                            "weight": {
                                "sqltype": "INTEGER NOT NULL",
                                "quoted": false
                            },
                            "job": {
                                "sqltype": "VARCHAR(50)",
                                "quoted": true
                            },
                            "datetime": {
                                "sqltype": "DATETIME",
                                "quoted": false
                            }
                        }
                    }
                },
                "serializers": {
                    "to": [
                        {
                            "type": "file",
                            "uri": "/home/tadeu/Desktop/dataset.sql"
                        }
                    ]
                }
            }
        }
    }



from pytest import fixture


def valid_specfication(format_type):
    return {
        'datasets': {
            'valid': {
                'fields': [
                    {'name': 'id',
                     'type': 'integer:sequence',
                     'generator': {
                         'start_at': 0
                     }},
                    {'name': 'name',
                     'type': 'pystr',
                     'generator': {
                         'length': 10
                     }},
                    {'name': 'age',
                     'type': 'pyint',
                     'generator': {
                         'start_at': 0,
                         'end_at': 120
                     }},
                    {'name': 'weight',
                     'type': 'pyfloat',
                     'generator': {
                         'start_at': 0.0,
                         'end_at': 500.0
                     }},
                    {'name': 'birth',
                     'type': 'date_between',
                     'generator': {
                         'start_date': '1980-01-01 00:00:00',
                         'end_date': '2020-01-01 00:00:00'
                     }},
                    {'name': 'hiring_at',
                     'type': 'timestamp:sequence',
                     'generator': {
                         'start_at': '1980-01-01 00:00:00'
                     }},
                ],
                'format': {
                    'type': format_type
                },
                'locale': 'pt_BR',
                'serializers': {
                    'to': [
                        {
                            'type': 'file',
                            'uri': ''
                        }
                    ]
                },
                'size': 100
            }
        }
    }


def invalid_specfication(format_type):
    return {
        'datasets': {
            'invalid': {
                'fields': [
                    {'name': 'id',
                     'type': 'integer:sequence'},
                    {'name': 'reference_id',
                     'type': 'integer:sequence',
                     'generator': {}},
                    {'name': 'level',
                     'type': 'integer',
                     'generator': {
                         'start_at': 2147483649,
                         'end_at': -2147483649
                     }},
                    {'name': 'rank',
                     'type': 'integer',
                     'generator': {
                         'start_at': 5,
                         'end_at': 4
                     }},
                    {'name': 'name',
                     'type': 'string'},
                    {'name': 'last_name',
                     'type': 'string',
                     'generator': {
                         'length': -1
                     }},
                    {'name': 'age',
                     'type': 'integer',
                     'generator': {}},
                    {'name': 'weight',
                     'type': 'float',
                     'generator': {}},
                    {'name': 'money',
                     'type': 'float',
                     'generator': {
                         'start_at': -2350738585072014e-308,
                         'end_at': 7976931348623157e+309
                     }},
                    {'name': 'taxes',
                     'type': 'float',
                     'generator': {
                         'start_at': 8.0,
                         'end_at': -1.0
                     }},
                    {'name': 'sex',
                     'type': None},
                    {'name': 'birth',
                     'type': 'timestamp'},
                    {'name': 'created_at',
                     'type': 'timestamp',
                     'generator': {
                         'start_at': 'heuheuhe',
                         'end_at': 'uheuhehu'
                     }},
                    {'name': 'updated_at',
                     'type': 'timestamp',
                     'generator': {
                         'start_at': '2020-01-01 00:00:00',
                         'end_at': '2010-01-01 00:00:00'
                     }},
                    {'name': 'hiring_at',
                     'type': 'timestamp:sequence'},
                    {'name': 'fired_at',
                     'type': 'timestamp:sequence',
                     'generator': {}},
                ],
                'format': {
                    'type': format_type
                },
                'locale': 'pt_BR',
                'serializers': {
                    'to': []
                },
                'size': '100'
            }
        }
    }


@fixture
def valid_csv_specification():
    return valid_specfication('csv')


@fixture
def valid_json_specification():
    return valid_specfication('json')


@fixture
def invalid_csv_specification():
    return invalid_specfication('csv')


@fixture
def invalid_json_specification():
    return invalid_specfication('json')


@fixture
def str_type_rule():
    return {'item': {'none': False, 'type': str}}


@fixture
def dict_type_rule():
    return {'item': {'none': False, 'type': dict}}


@fixture
def invalid_type_sample():
    return {'item': 15}


@fixture
def fixed_index_rule():
    return {'items.[1].age': {'none': False, 'type': int}}


@fixture
def collection_sample():
    return {'items': [{'age': '5'}, {'age': '7'}, {'age': '15'}]}


@fixture
def fixed_key_rule():
    return {'items.{second}.age': {'none': False, 'type': int}}


@fixture
def dictionary_sample():
    return {
        'items': {
            'first': {'age': '5'},
            'second': {'age': '7'},
            'third': {'age': '15'}
        }
    }


@fixture
def csv_format_sample():
    return {'type': 'csv'}


@fixture
def json_format_sample():
    return {'type': 'json-array'}


@fixture
def sql_format_sample():
    return {'type': 'sql'}


@fixture
def unknown_format_sample():
    return {'type': 'unknown'}


@fixture
def file_writer_sample():
    return {'type': 'file'}


@fixture
def mysql_writer_sample():
    return {'type': 'sql',
            'options': {
                'method': 'cli',
                'engine': 'mysql'
            }}


@fixture
def postgresql_writer_sample():
    return {'type': 'sql',
            'options': {
                'method': 'cli',
                'engine': 'postgres'
            }}


@fixture
def s3_writer_sample():
    return {'type': 's3'}


@fixture
def s3_url_writer_sample():
    return {'type': 's3-url'}


@fixture
def gcs_writer_sample():
    return {'type': 'gcs'}


@fixture
def gcs_url_writer_sample():
    return {'type': 'gcs-url'}


@fixture
def unknown_writer_sample():
    return {'type': 'unknown'}


@fixture
def invalid_s3_bucket():
    return {
        "to": [{
            "type": "s3",
            "options": {
                "key": "key"
            }}
        ]
    }


@fixture
def invalid_s3_key():
    return {
        "to": [{
            "type": "s3",
            "options": {
                "bucket": "bucket"
            }}
        ]
    }


@fixture
def expected_values_rule():
    return {'items.{*}.age': {'none': False,
                              'type': str,
                              'values': ['3', '9']}}


def custom_rule_method(value):
    if value != 3:
        raise ValueError("Test value is different, expected: 3")


@fixture
def custom_rule():
    return {'items.{*}.age': {'none': False,
                              'type': str,
                              'custom': [custom_rule_method]}}



from pandas import DataFrame


class JsonFormatter(object):
    """Class that receive pandas dataframe
    and write it down in Json format
    """
    key = 'json-array'

    def __init__(self, specification={}):
        self.default = {'orient': 'records',
                        'date_unit': 's'}
        self.specification = specification

    @staticmethod
    def rules():
        return {'required': {},
                'optional': {
                    'options.orient': {'none': False, 'type': str},
                    'options.index': {'none': False, 'type': bool}
                }}

    def format(self, dataframe: DataFrame, path_or_buffer) -> None:
        """Format dataframe to json.

        Keyword arguments:
         - dataframe - pandas.Dataframe: dataframe containing the records
        """
        parameters = self.default
        options = self.specification.get('options', {})
        options['indent'] = 2 if options.get("indent") == "pretty" else 0
        parameters.update(options)
        if dataframe.shape[0] > 0:
            dataframe.to_json(
                path_or_buf=path_or_buffer, force_ascii=False, **parameters)



import pandas as pd

from sqlalchemy import create_engine

from src.lib.formatters.sql import SQLFormatter
from src.lib.writers.databases.mysql import MysqlDirectDatabaseWriter
from src.tests.lib.writers.fixtures import *  # noqa: F403, F401
import os
import uuid


class TestSQLWriter(object):
    """Unit-test for SQLWriter class"""

    def test_writting_sql_table_without_index(self,
                                              mocker,
                                              pandas_dataframe_with_data,
                                              specification,
                                              sql_specification_format,
                                              sql_specification_writer):
        expected = "123456789"
        table_name = "mytable"
        index_flag = False
        formatter_spec = sql_specification_format(table_name=table_name,
                                                  index=index_flag)

        formatter = SQLFormatter(specification=formatter_spec)
        writer = MysqlDirectDatabaseWriter(
            formatter=formatter,
            specification=sql_specification_writer
        )

        db_engine = create_engine('sqlite:///:memory:')

        mock = mocker.patch('getpass._raw_input')
        mock.return_value = expected

        mock_ = mocker.patch.object(MysqlDirectDatabaseWriter, 'engine')
        mock_.return_value = db_engine

        writer.before_write()
        writer.write(dataframe=pandas_dataframe_with_data)

        dataframe_from_sql = pd.read_sql_table(
            table_name='mytable', con=db_engine)
        assert dataframe_from_sql.equals(pandas_dataframe_with_data)
        mock_.assert_called()

    def test_writting_sql_table_with_index(self,
                                           mocker,
                                           pandas_dataframe_with_data,
                                           specification,
                                           sql_specification_format,
                                           sql_specification_writer):
        expected = "123456789"
        table_name = "mytable"
        index_flag = True
        index_label = "myindexlabel"
        formatter_spec = sql_specification_format(table_name=table_name,
                                                  index=index_flag,
                                                  index_label=index_label)

        formatter = SQLFormatter(specification=formatter_spec)
        writer = MysqlDirectDatabaseWriter(
            formatter=formatter,
            specification=sql_specification_writer
        )

        db_engine = create_engine('sqlite:///:memory:')

        mock = mocker.patch('getpass._raw_input')
        mock.return_value = expected

        mock_ = mocker.patch.object(MysqlDirectDatabaseWriter, 'engine')
        mock_.return_value = db_engine

        writer.before_write()
        writer.write(dataframe=pandas_dataframe_with_data)

        pandas_dataframe_with_data.index.name = index_label
        pandas_dataframe_with_data = \
            pandas_dataframe_with_data.reset_index(level=0)

        dataframe_from_sql = pd.read_sql_table(
            table_name='mytable', con=db_engine)
        assert dataframe_from_sql.equals(pandas_dataframe_with_data)
        mock_.assert_called()

    def test_writing_sql_script(self, sql_specification_format,
                                pandas_dataframe_with_data):
        table_name = "mytable"
        index_flag = True
        index_label = "myindexlabel"
        uri = uuid.uuid4().hex
        formatter_spec = sql_specification_format(table_name=table_name,
                                                  index=index_flag,
                                                  index_label=index_label)
        formatter = SQLFormatter(specification=formatter_spec)
        formatter.format(dataframe=pandas_dataframe_with_data,
                         path_or_buffer=uri)

        assert os.path.exists(uri)
        with open(uri, 'r') as f:
            assert len(f.read()) > 0
        os.remove(uri)



from pytest import raises

from src.lib.config.serializers import SerializersConfiguration
from src.tests.lib.config.fixtures import *  # noqa: F403, F401


class TestSerializersConfiguration(object):
    """Unit-test of TestSerializersConfiguration class"""

    def test_file(self, file_writer_sample):
        rules = SerializersConfiguration.get_rules(file_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_mysql(self, mysql_writer_sample):
        rules = SerializersConfiguration.get_rules(mysql_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_postgresql(self, postgresql_writer_sample):
        rules = SerializersConfiguration.get_rules(postgresql_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_s3(self, s3_writer_sample):
        rules = SerializersConfiguration.get_rules(s3_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_s3_url(self, s3_url_writer_sample):
        rules = SerializersConfiguration.get_rules(s3_url_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_gcs(self, gcs_writer_sample):
        rules = SerializersConfiguration.get_rules(gcs_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_gcs_url(self, gcs_url_writer_sample):
        rules = SerializersConfiguration.get_rules(gcs_url_writer_sample)

        assert rules is not None
        assert isinstance(rules, dict) is True

    def test_unknown(self, unknown_writer_sample):
        with raises(ValueError):
            SerializersConfiguration.get_rules(unknown_writer_sample)



from pandas import DataFrame
from sqlalchemy.engine import create_engine

from src.lib.postgresql.psql import PostgresSqlPsql
from src.lib.writers.databases import DatabaseWriter


class PostgreSqlClientDatabaseWriter(DatabaseWriter):
    key = 'sql'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.specification = specification

    def write(self, dataframe: DataFrame):
        params = self.specification \
                     .get('options')

        sql = self.formatter \
                  .format(dataframe=dataframe,
                          path_or_buffer=None)
        if sql is not None:
            psql = PostgresSqlPsql(host=params['host'],
                                   port=params.get('port', 5432),
                                   database=params['database'],
                                   user=params['username'],
                                   password=params['password'])
            for query in sql.split(';'):
                if len(query) < 6:
                    continue
                psql.execute_query(query)

        return "Inserted with success"


class PostgresDirectDatabaseWriter(DatabaseWriter):
    """Class that receive pandas dataframe
    and write it on a SGDB
    """
    key = 'sql'

    def __init__(self, formatter, specification):
        self.formatter = formatter
        self.default = {
            'schema': 'public'
        }
        self.specification = specification

    def engine(self, params):
        return create_engine('postgres://'
                             f"{params['username']}:"
                             f"{params['password']}@"
                             f"{params['host']}:"
                             f"{params.get('port', 5432)}/"
                             f"{params['database']}")

    def write(self, dataframe: DataFrame) -> None:
        """Write all dataframe on a DataBase Table.

        Keyword arguments:
            - dataframe - pandas.Dataframe: dataframe containing the records
        """
        parameters = self.default
        options = self.specification.get('options', {})
        parameters.update(options)

        connection = self.engine(parameters)

        self.formatter.format(dataframe=dataframe,
                              path_or_buffer=connection,
                              method='direct',
                              schema=parameters['schema'])
        return "Inserted with success"



from os import remove
from os.path import exists, isfile

from src.lib.formatters.json import JsonFormatter
from src.tests.lib.writers.fixtures import *  # noqa: F403, F401


class TestJsonWriter:
    """Unit-test of JsonWriter class"""

    def test_writting_dataframe_with_records(self,
                                             pandas_dataframe_with_data,
                                             specification):
        test_file_path = specification['uri']
        json_writer = JsonFormatter()
        json_writer.format(dataframe=pandas_dataframe_with_data,
                           path_or_buffer=test_file_path)
        assert exists(test_file_path) is True
        assert isfile(test_file_path) is True

        with open(test_file_path) as json_file:
            expected = pandas_dataframe_with_data.to_json(orient="records")

            assert json_file.read() == expected

        remove(test_file_path)
        assert exists(test_file_path) is False

    def test_writting_dataframe_with_records_orient_columns(
            self,
            pandas_dataframe_with_data,
            specification):
        test_file_path = specification['uri']
        json_writer = JsonFormatter(specification={
            "options":
                {"orient": "columns"}
            }
        )
        json_writer.format(dataframe=pandas_dataframe_with_data,
                           path_or_buffer=test_file_path)
        assert exists(test_file_path) is True
        assert isfile(test_file_path) is True

        with open(test_file_path) as json_file:
            expected = pandas_dataframe_with_data.to_json(orient="columns")

            assert json_file.read() == expected

        remove(test_file_path)
        assert exists(test_file_path) is False

    def test_writting_dataframe_with_records_orient_records(
            self,
            pandas_dataframe_with_data,
            specification):
        test_file_path = specification['uri']
        json_writer = JsonFormatter(specification={
            "options":
                {"orient": "records"}
            }
        )
        json_writer.format(dataframe=pandas_dataframe_with_data,
                           path_or_buffer=test_file_path)
        assert exists(test_file_path) is True
        assert isfile(test_file_path) is True

        with open(test_file_path) as json_file:
            expected = pandas_dataframe_with_data.to_json(orient="records")

            assert json_file.read() == expected

        remove(test_file_path)
        assert exists(test_file_path) is False

    def test_writting_dataframe_without_records(self,
                                                pandas_dataframe_without_data,
                                                specification):
        test_file_path = specification['uri']
        json_writer = JsonFormatter()
        json_writer.format(dataframe=pandas_dataframe_without_data,
                           path_or_buffer=test_file_path)
        assert exists(test_file_path) is False
        assert isfile(test_file_path) is False



from src.lib.writers.remotes.gcs import GCSRemoteWriter
from src.lib.formatters import formatters
from src.tests.lib.writers.remotes.fixtures import *  # noqa: F403, F401
from csv import reader
import gcsfs
from io import StringIO
spec_options = {
        "options": {
            "bucket": "ggnerator",
            "key": "path/to/file.csv"
        }
    }


class TestGcsWriter:
    def test_write_dataframe_csv(self, mocker,
                                 pandas_dataframe_with_data,
                                 specification_gcs):
        buffer = StringIO()
        mock = mocker.patch.object(gcsfs, 'GCSFileSystem')
        mock.return_value.open.return_value = buffer

        csv_formatter = formatters['csv']({"options": {"header": True}})
        gcs = GCSRemoteWriter(formatter=csv_formatter,
                              specification=specification_gcs)
        gcs.write(pandas_dataframe_with_data)
        expected = pandas_dataframe_with_data.to_dict(orient="list")

        buffer.seek(0)
        csv_reader = reader(buffer.readlines(), delimiter=',')
        csv_reader = [x for x in csv_reader]
        headers = csv_reader[0]
        rows = csv_reader[1:len(csv_reader)]
        csv_reader = {headers[index]: [row[index]
                      for row in rows]
                      for index in range(0, len(headers))}
        assert csv_reader == expected

    def test_write_dataframe_json(self, mocker,
                                  pandas_dataframe_with_data,
                                  specification_gcs):
        buffer = StringIO()
        mock = mocker.patch.object(gcsfs, 'GCSFileSystem')
        mock.return_value.open.return_value = buffer

        json_formatter = formatters['json-array'](
            {"options": {"orient": "records"}})
        gcs = GCSRemoteWriter(formatter=json_formatter,
                              specification=specification_gcs)
        gcs.write(pandas_dataframe_with_data)
        expected = pandas_dataframe_with_data.to_json(orient="records")

        buffer.seek(0)
        assert buffer.read() == expected



import os

from click.testing import CliRunner


from src.cli.commands import execute
from src.generators.handler import GeneratorsHandler
from src.generators.datatypes import Metadata
from src.tests.cli.cli_fixture import *  # noqa: F403, F401


class TestCliExecutor:
    def test_cli_executor_file(self, mocker, generate_files):
        runner = CliRunner()
        mock = mocker.patch.object(GeneratorsHandler, 'generate')
        mock.return_value = [("a", "b", "c")]
        mock_ = mocker.patch.object(GeneratorsHandler,
                                    'valid_specification_dataset')
        mock_.return_value = {}

        with runner.isolated_filesystem():
            with open('hello.json', 'w') as f:
                f.write("{{}}")
            result = runner.invoke(execute, ['generate',
                                             '--spec',
                                             'hello.json'])
            assert result.exit_code == 0

    def test_cli_executor_folder(self, mocker, generate_files):
        runner = CliRunner()
        mock = mocker.patch.object(GeneratorsHandler, 'generate')
        mock.return_value = [("a", "b", "c")]
        mock_ = mocker.patch.object(GeneratorsHandler,
                                    'valid_specification_dataset')
        mock_.return_value = {}

        with runner.isolated_filesystem():
            os.mkdir('dir/')
            result = runner.invoke(execute, ['generate',
                                             '--spec',
                                             'dir/'])
            assert result.exit_code != 0

    def test_cli_executor_file_not_found(self, mocker, generate_files):
        runner = CliRunner()
        mock = mocker.patch.object(GeneratorsHandler, 'generate')
        mock.return_value = [("a", "b", "c")]
        mock_ = mocker.patch.object(GeneratorsHandler,
                                    'valid_specification_dataset')
        mock_.return_value = {}

        with runner.isolated_filesystem():
            result = runner.invoke(execute, ['generate',
                                             '--spec',
                                             'hello.json'])
            assert result.exit_code != 0

    def test_list_generators(self, mocker, generators_map):
        runner = CliRunner()
        mock = mocker.patch.object(Metadata, 'get_generators_map')
        mock.return_value = generators_map

        with runner.isolated_filesystem():
            result = runner.invoke(execute, ['list-generators'])
            assert result.exit_code == 0

    def test_sample_generators(self):
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(execute, ['generate-sample',
                                             '--type',
                                             'integer'])
            assert result.exit_code == 0

    def test_sample_generators_invalid_type(self):
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(execute, ['generate-sample',
                                             '--type',
                                             'abcdfg'])
            assert "Type not found" in result.stdout



import pytz

from datetime import datetime, timedelta
from dateutil.parser import isoparse
from typing import List


class TimestampSequenceType:
    key = 'timestamp:sequence'
    namespace = 'basic_type'
    optional_arguments = False

    def __init__(self, start_at: str,
                 datepart: str = "second", tz: str = "UTC", step: int = 1):
        """
        step must be
        second | minute | hour | day | month | year
        default: second
        """
        self.start_date = self.__parse_to_datetime(start_at)
        self.datepart = datepart
        self.step = step
        self.datepart_in_seconds = {
            "year": 365.25*24*60*60,
            "month": 365.25/12*24*60*60,
            "day": 24*60*60,
            "hour": 60*60,
            "minute": 60,
            "second": 1
        }
        self.time_zone = pytz.timezone(tz)

    def __parse_to_datetime(self, iso_format_date) -> datetime:
        return isoparse(iso_format_date)

    def __generate_next(self, delta) -> datetime:
        return (self.start_date + self.__delta_step(delta))\
            .astimezone(tz=self.time_zone)

    def __delta_step(self, delta):
        seconds = self.datepart_in_seconds[self.datepart]
        return timedelta(seconds=seconds * delta * self.step)

    def generate(self, num_of_records: int = 5) -> List[str]:
        return [dt.isoformat()
                for dt in self.generate_records(num_of_records=num_of_records)]

    def generate_records(self,
                         num_of_records,
                         progress=None) -> List[datetime]:
        results = []
        for delta in range(0, num_of_records):
            results.append(self.__generate_next(delta))
            progress and progress()

        return results

    @staticmethod
    def rules():
        def validate(value):
            try:
                isoparse(value)
            except ValueError:
                raise ValueError(f"Timestamp field `{value}` "
                                 "has a wrong format")
        return {
            'required': {
                'generator.start_at': {
                    'none': False,
                    'type': str,
                    'custom': [validate]
                }
            },
            'optional': {}
        }



from re import match


class ConfigurationInpector(object):

    def __validate_none(self, value, expected):
        return expected.get('none') is False and value is None

    def __validate_type(self, value, expected):
        return not isinstance(value, expected['type'])

    def __validate_values(self, value, expected_values):
        return value not in expected_values

    def __validate(self,
                   configuration: dict,
                   rule: str,
                   expected: dict,
                   optional: bool,
                   trace):
        keys = rule.split('.')
        key = keys[0]

        if self.__validate_none(value=configuration.get(key),
                                expected=expected) and \
           optional is False:
            raise ValueError(f"JSON Key `{trace}` can not "
                             "be `None`, `null` or empty")
        elif self.__validate_none(value=configuration.get(key),
                                  expected=expected) and optional is True:
            return

        if len(keys) > 1:
            return

        value = configuration.get(key)
        if self.__validate_type(value=value, expected=expected):
            expected_type = expected['type']
            if expected_type == dict:
                expected_type = type(expected_type)
            raise ValueError(f"JSON Key `{trace}` must be "
                             f"of type `{expected_type}`")

        if 'values' in expected:
            expected_values = expected.get('values', None)
            if self.__validate_values(value=value,
                                      expected_values=expected_values):
                raise ValueError(f"JSON Key `{trace}` is invalid, these are "
                                 f"the possible options: `{expected_values}`")

        custom_rules = expected.get('custom', [])
        if len(custom_rules) > 0:
            for custom in custom_rules:
                custom(value)

    def inspect_rule(self,
                     configuration: dict,
                     rule: str,
                     expected: dict,
                     optional: bool = False,
                     trace=""):
        current_key = rule.split('.')[0]
        partial_rule = rule.replace(f'{current_key}.', "")

        if match(r'[{\[].*[}\]]', current_key):
            if current_key[1] == '*':  # Iterable
                if len(configuration) <= 0:
                    raise ValueError(f"Key '{trace}{current_key}' has "
                                     "no valid values")
                if current_key[0] == '[':  # List
                    for index in range(0, len(configuration)):
                        item = configuration[index]
                        key_trace = f"{trace}{index}."
                        self.inspect_rule(configuration=item,
                                          expected=expected,
                                          rule=partial_rule,
                                          optional=optional,
                                          trace=key_trace)
                    return
                else:  # Dict
                    for key in configuration.keys():
                        key_trace = f"{trace}{key}."
                        self.inspect_rule(
                            configuration=configuration[key],
                            expected=expected,
                            rule=partial_rule,
                            optional=optional,
                            trace=key_trace
                        )
                    return
            else:  # Fixed Index
                index = current_key[1:-1]
                key_trace = f"{trace}{index}"
                if current_key[0] == '[':  # List
                    index = int(index)
                    return self.inspect_rule(
                        configuration=configuration[index],
                        expected=expected,
                        rule=partial_rule,
                        optional=optional,
                        trace=key_trace
                    )
                else:  # Dict
                    return self.inspect_rule(
                        configuration=configuration[index],
                        expected=expected,
                        rule=partial_rule,
                        optional=optional,
                        trace=key_trace
                    )

        self.__validate(configuration=configuration,
                        rule=rule,
                        expected=expected,
                        optional=optional,
                        trace=f"{trace}{current_key}")
        if '.' not in rule or \
           (current_key not in configuration and optional is True):
            return
        return self.inspect_rule(configuration=configuration[current_key],
                                 expected=expected,
                                 rule=partial_rule,
                                 optional=optional,
                                 trace=f"{trace}{current_key}.")

    def inspect_rules(self, rules, configuration, optional=False, trace=""):
        for rule_key in rules.keys():
            self.inspect_rule(rule=rule_key,
                              expected=rules[rule_key],
                              configuration=configuration,
                              optional=optional,
                              trace=trace)



from os import getenv
from src.lib.shell import Shell

POSTGRES_CLI_BINPATH = getenv("POSTGRES_CLI_BINPATH")


class PostgresSqlPsql(object):
    connection = ("host={host} port={port} user={user} "
                  "password={password} dbname={database}")

    def __init__(self, host, port, database, user, password):
        self.shell = Shell()
        self.base_command = self.connection\
                                .format(host=host,
                                        port=port,
                                        user=user,
                                        password=password,
                                        database=database)

    def execute_query(self, query):
        nl = "\n"
        default_psql = [POSTGRES_CLI_BINPATH or '/usr/bin/env', 'psql']

        command = [
            *default_psql,
            '-c',
            f'{query.replace(nl, "")}',
            self.base_command
        ]
        output = self.shell \
                     .execute(command=command)

        return output



from pytest import fixture


@fixture
def simple_specification():
    return {'size': 100,
            'locale': 'pt_BR',
            'fields': [{'name': 'id',
                        'type': 'integer:sequence',
                        'generator': {},
                        'expected': 'int64'},
                       {'name': 'text',
                        'type': 'pystr',
                        'generator': {},
                        'expected': 'object'}]}


@fixture
def argumented_specification():
    return {'size': 200,
            'locale': 'pt_BR',
            'fields': [{'name': 'id',
                        'type': 'integer:sequence',
                        'generator': {'start_at': 10,
                                      'step': 2},
                        'expected': 'int64'},
                       {'name': 'date',
                        'type': 'date_time_between',
                        'generator': {'start_date': "-10d",
                                      'end_date':  '+1y'},
                        'expected': 'datetime64[ns]'}]}


def sample(type, expected):
    return {'size': 200,
            'locale': 'pt_BR',
            'fields': [{'name': 'id',
                        'type': type,
                        'generator': {},
                        'expected': expected}]}


def dataframe_sample(type, expected):
    return {
        'datasets': {
            'sample': {
                'size': 200,
                'locale': 'pt_BR',
                'fields': [
                    {
                        'name': 'id',
                        'type': type,
                        'generator': {},
                        'expected': expected
                    }
                ],
                'format': {'type': 'csv'},
                'serializers': {
                    'to': [
                        {
                            'type': 'file',
                            'uri': ''
                        }
                    ]
                }
            }
        }
    }


def sample_timestamp(type, expected):
    return {'size': 200,
            'locale': 'pt_BR',
            'fields': [{'name': 'id',
                        'type': type,
                        'generator': {
                            'start_date': '-1d',
                            'end_date':  '+1d'
                        },
                        'expected': expected}]}


def sample_timestamp_sequence(type, expected):
    return {'size': 200,
            'locale': 'pt_BR',
            'fields': [{'name': 'id',
                        'type': type,
                        'generator': {
                            'start_at': '2019-01-01T01:00:00Z'
                        },
                        'expected': expected}]}


@fixture
def integer_specification():
    return sample('pyint', 'int64')


@fixture
def bool_specification():
    return dataframe_sample('boolean', 'bool')


@fixture
def char_specification():
    return sample('pystr', 'object')


@fixture
def name_specification():
    return sample('name', 'object')


@fixture
def float_specification():
    return sample('pyfloat', 'float64')


@fixture
def timestamp_sequence_specification():
    return sample_timestamp_sequence('timestamp:sequence', 'datetime64[ns]')


@fixture
def integer_sequence_specification():
    return sample('integer:sequence', 'int64')


@fixture
def string_specification():
    return sample('pystr', 'object')


@fixture
def timestamp_specification():
    return sample_timestamp('date_time_between', 'datetime64[ns]')


@fixture
def valid_specification():
    return {
        "datasets": {
            "teste3": {
                "fields": [
                    {
                        "type": "integer:sequence",
                        "name": "id",
                        "generator": {
                            "start_at": 1
                        }
                    }
                ],
                "size": 100,
                "locale": "pt_BR",
                "format": {
                    "type": "csv"
                },
                "serializers": {
                    "to": [
                        {
                            "type": "file",
                            "uri": "/tmp/teste.csv"
                        }
                    ]
                }
            }
        }
    }


@fixture
def no_datasets_specification():
    return {}


@fixture
def invalid_no_ids_dataset():
    return {"datasets": {}}


@fixture
def invalid_no_size_dataset():
    return {"datasets": {
        "$id": {
            "fields": [],
            "locale": "",
            "format": {},
            "serializers": {}
        }
    }
    }


@fixture
def invalid_no_locale_dataset():
    return {"datasets": {
        "$id": {
            "fields": [],
            "size": 10,
            "format": {},
            "serializers": {}
        }
    }
    }


@fixture
def invalid_no_size_dryrun():
    return {"datasets": {
        "$id": {
            "fields": [],
            "format": {},
            "serializers": {}
        }
    }
    }


@fixture
def invalid_dataset_specification():
    return {"datasets": {
        "teste3": {
            "fields": [{"type": "integer:sequence",
                                "name": "id",
                                "generator": {"start_at": 1}}],
            "size": "",
                    "locale": "pt_BR",
                    "format": {"type": "csv"},
                    "serializers": {
                        "to": [{"type": "file",
                                "uri": "/tmp/teste.csv"}]}}}}


@fixture
def valid_dryrun():
    return {"size": 10,
            "locale": "pt_BR",
            "fields": [{
                "name": "code",
                "type": "integer:sequence",
                "generator": {"start_at": 10}
            }]
            }


@fixture
def invalid_dateformat():
    return {"datasets": {
        "teste3": {
            "fields": [{"type": "integer:sequence",
                                "name": "id",
                                "generator": {"start_at": 1}},
                       {"type": "date_time",
                                "name": "created_at",
                                "generator": {}}],
            "size": 1000,
            "locale": "pt_BR",
            "format": {"type": "csv",
                       "options": {
                           "header": True,
                           "sep": ","
                       }},
            "serializers": {
                "to": [{"type": "file",
                        "uri": "/tmp/teste.csv"}]}}}}


@fixture
def malformed_json():
    return "[]]"


@fixture
def unknown_type_spec():
    return {"datasets": {
        "teste3": {
            "fields": [{"type": "timestamp",
                        "name": "created_at",
                        "generator": {}}],
            "size": 1000,
            "locale": "pt_BR",
            "format": {"type": "csv",
                       "options": {
                           "header": True,
                           "sep": ","
                       }},
            "serializers": {
                "to": [{"type": "file",
                        "uri": "/tmp/teste.csv"}]}}}}


@fixture
def valid_spec_for_replace_rules():
    false = False
    true = True

    return {
        "datasets": {
            "sample": {
                "size": 10,
                "locale": "pt_BR",
                "fields": [
                    {
                        "type": "integer:sequence",
                        "name": "id",
                        "generator": {
                            "start_at": 1
                        }
                    },
                    {
                        "type": "name",
                        "name": "name",
                        "generator": {}
                    },
                    {
                        "type": "pyint",
                        "name": "age",
                        "generator": {
                            "max_value": 120
                        }
                    },
                    {
                        "type": "pyfloat",
                        "name": "weight",
                        "generator": {
                            "positive": false,
                            "min_value": 0,
                            "max_value": 250
                        }
                    },
                    {
                        "type": "job",
                        "name": "job",
                        "generator": {}
                    },
                    {
                        "type": "future_datetime",
                        "name": "datetime",
                        "generator": {}
                    }
                ],
                "format": {
                    "type": "sql",
                    "options": {
                        "table_name": "My_table",
                        "mode": "replace",
                        "schema": {
                            "id": {
                                "sqltype": "INTEGER NOT NULL",
                                "quoted": false
                            },
                            "name": {
                                "sqltype": "VARCHAR(50)",
                                "quoted": true
                            },
                            "age": {
                                "sqltype": "INTEGER NOT NULL",
                                "quoted": false
                            },
                            "weight": {
                                "sqltype": "INTEGER NOT NULL",
                                "quoted": false
                            },
                            "job": {
                                "sqltype": "VARCHAR(50)",
                                "quoted": true
                            },
                            "datetime": {
                                "sqltype": "DATETIME",
                                "quoted": false
                            }
                        }
                    }
                },
                "serializers": {
                    "to": [
                        {
                            "type": "file",
                            "uri": "/home/tadeu/Desktop/dataset.sql"
                        }
                    ]
                }
            }
        }
    }


@fixture
def invalid_spec_for_replace_rules_without_schema():
    false = False

    return {
        "datasets": {
            "sample": {
                "size": 10,
                "locale": "pt_BR",
                "fields": [
                    {
                        "type": "integer:sequence",
                        "name": "id",
                        "generator": {
                            "start_at": 1
                        }
                    },
                    {
                        "type": "name",
                        "name": "name",
                        "generator": {}
                    },
                    {
                        "type": "pyint",
                        "name": "age",
                        "generator": {
                            "max_value": 120
                        }
                    },
                    {
                        "type": "pyfloat",
                        "name": "weight",
                        "generator": {
                            "positive": false,
                            "min_value": 0,
                            "max_value": 250
                        }
                    },
                    {
                        "type": "job",
                        "name": "job",
                        "generator": {}
                    },
                    {
                        "type": "future_datetime",
                        "name": "datetime",
                        "generator": {}
                    }
                ],
                "format": {
                    "type": "sql",
                    "options": {
                        "table_name": "My_table",
                        "mode": "replace"
                    }
                },
                "serializers": {
                    "to": [
                        {
                            "type": "file",
                            "uri": "/home/tadeu/Desktop/dataset.sql"
                        }
                    ]
                }
            }
        }
    }


@fixture
def invalid_spec_for_replace_rules_without_sqltype():
    false = False

    return {
        "datasets": {
            "sample": {
                "size": 10,
                "locale": "pt_BR",
                "fields": [
                    {
                        "type": "integer:sequence",
                        "name": "id",
                        "generator": {
                            "start_at": 1
                        }
                    },
                    {
                        "type": "name",
                        "name": "name",
                        "generator": {}
                    },
                    {
                        "type": "pyint",
                        "name": "age",
                        "generator": {
                            "max_value": 120
                        }
                    },
                    {
                        "type": "pyfloat",
                        "name": "weight",
                        "generator": {
                            "positive": false,
                            "min_value": 0,
                            "max_value": 250
                        }
                    },
                    {
                        "type": "job",
                        "name": "job",
                        "generator": {}
                    },
                    {
                        "type": "future_datetime",
                        "name": "datetime",
                        "generator": {}
                    }
                ],
                "format": {
                    "type": "sql",
                    "options": {
                        "table_name": "My_table",
                        "mode": "replace",
                        "schema": {
                            "id": {
                                "quoted": false
                            }
                        }
                    }
                },
                "serializers": {
                    "to": [
                        {
                            "type": "file",
                            "uri": "/home/tadeu/Desktop/dataset.sql"
                        }
                    ]
                }
            }
        }
    }



import random

from src.generators.datatypes.sequence import SequenceType
from src.tests.generators.datatypes.generators_fixtures import *  # noqa: F403, F401, E501

TEST_NUM_OF_RECORDS = random.randint(1, 10000)
TEST_RANDOM_STEP = random.randint(1, 1000)


class TestSequenceGenerator:

    def test_num_of_records(self, generate):
        records = generate(SequenceType,
                           TEST_NUM_OF_RECORDS,
                           step=TEST_RANDOM_STEP)
        assert len(records) == TEST_NUM_OF_RECORDS

    def test_all_records_integer(self, generate):
        records = generate(SequenceType,
                           TEST_NUM_OF_RECORDS,
                           step=TEST_RANDOM_STEP)
        for record in records:
            assert isinstance(record, int)

    def test_step_sequence(self, generate):
        records = generate(SequenceType,
                           TEST_NUM_OF_RECORDS,
                           step=TEST_RANDOM_STEP)
        end_sublist = records[1:]
        start_sublist = records[:-1]
        differences_list = \
            [second - first
             for (first, second) in zip(start_sublist, end_sublist)]
        for difference in differences_list:
            assert difference == TEST_RANDOM_STEP



from pytest import raises

from src.lib.config import general_rules
from src.lib.config.inspector import ConfigurationInpector
from src.tests.lib.config.fixtures import *  # noqa: F403, F401


class TestConfigurationInspector(object):
    """Unit-test of TestConfigurationInspector class"""

    def test_valid_csv_configuration(self, valid_csv_specification):
        inspector = ConfigurationInpector()
        inspect = inspector. \
            inspect_rules(rules=general_rules,
                          configuration=valid_csv_specification)
        assert inspect is None

    def test_valid_json_configuration(self, valid_json_specification):
        inspector = ConfigurationInpector()
        inspect = inspector. \
            inspect_rules(rules=general_rules,
                          configuration=valid_json_specification)
        assert inspect is None

    def test_invalid_csv_configuration(self, invalid_csv_specification):
        inspector = ConfigurationInpector()
        with raises(ValueError):
            inspector.inspect_rules(rules=general_rules,
                                    configuration=invalid_csv_specification)

    def test_invalid_json_configuration(self, invalid_json_specification):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=general_rules,
                                    configuration=invalid_json_specification)

    def test_invalid_type(self, str_type_rule, invalid_type_sample):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=str_type_rule,
                                    configuration=invalid_type_sample)

    def test_invalid_dict_type(self, dict_type_rule, invalid_type_sample):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=dict_type_rule,
                                    configuration=invalid_type_sample)

    def test_fixed_index_rule(self, fixed_index_rule, collection_sample):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=fixed_index_rule,
                                    configuration=collection_sample)

    def test_fixed_key_rule(self, fixed_key_rule, dictionary_sample):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=fixed_key_rule,
                                    configuration=dictionary_sample)

    def test_expected_values_rule(self,
                                  expected_values_rule,
                                  dictionary_sample):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=expected_values_rule,
                                    configuration=dictionary_sample)

    def test_custom_rule(self, custom_rule, dictionary_sample):
        inspector = ConfigurationInpector()

        with raises(ValueError):
            inspector.inspect_rules(rules=custom_rule,
                                    configuration=dictionary_sample)



from pytest import fixture


def generate_records(type_generator, num_of_records):
    return type_generator.generate_records(num_of_records)


@fixture
def generate():
    def _generate(DataType, num_of_records, **kwargs):
        return generate_records(DataType(**kwargs), num_of_records)
    return _generate



import json
import pytest

from mock import mock_open, patch
from os import remove
from pandas import DataFrame

from src.generators.basehandler import BaseHandler
from src.generators.dryrunhandler import DryRunHandler
from src.tests.generators.handler_fixtures import *  # noqa: F403, F401


class TestDryRunHandler(object):
    """Unit-test of DryRunHandler class"""

    def test_simple_specification_dataframe(self,
                                            mocker,
                                            simple_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = simple_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in simple_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_argumented_specification_dataframe(self,
                                                mocker,
                                                argumented_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = argumented_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in argumented_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_integer_dataframe(self, mocker, integer_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = integer_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in integer_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_bool_dataframe(self, mocker, bool_specification):
        with patch('builtins.open',
                   mock_open(read_data=json.dumps(bool_specification))):
            dataset = bool_specification['datasets']['sample']
            expected_size = 10
            expected_fields = dataset['fields']

            handler = DryRunHandler({"config_file": None})
            dataframe = handler.generate_dryrun(dataset)

            assert isinstance(dataframe, DataFrame) is True
            assert dataframe.shape[0] == expected_size
            for field in expected_fields:
                assert dataframe[field['name']].dtype.name == field['expected']

    def test_char_dataframe(self, mocker, char_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = char_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in char_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_float_dataframe(self, mocker, float_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = float_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in float_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_integer_sequence_dataframe(self,
                                        mocker,
                                        integer_sequence_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = integer_sequence_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in integer_sequence_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_timestamp_sequence_dataframe(self,
                                          mocker,
                                          timestamp_sequence_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = timestamp_sequence_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in timestamp_sequence_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_string_dataframe(self, mocker, string_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = string_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in string_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_timestamp_dataframe(self, mocker, timestamp_specification):
        mock = mocker.patch \
                     .object(DryRunHandler, 'valid_specification_dryrun')
        mock.return_value = timestamp_specification
        handler = DryRunHandler({'config_file': None})
        specification = handler.valid_specification_dryrun()
        dataframe = handler.generate_dryrun(specification)

        assert isinstance(dataframe, DataFrame) is True
        assert dataframe.shape[0] == 10
        for field in timestamp_specification['fields']:
            assert dataframe[field['name']].dtype.name == field['expected']

    def test_dryrunhandler_valid(self, valid_specification):
        with open('valid_spec.json', 'w') as f:
            json.dump(valid_specification, f)
        handler = DryRunHandler({'config_file': 'valid_spec.json'})
        config = handler.valid_specification_dryrun()
        assert valid_specification == config
        remove('valid_spec.json')

    def test_dryrunhandler_no_dataset_ids(self, invalid_no_ids_dataset):
        with open('invalid_spec.json', 'w') as f:
            json.dump(invalid_no_ids_dataset, f)
        with pytest.raises(ValueError):
            handler = DryRunHandler({'config_file': 'invalid_spec.json'})
            handler.valid_specification_dryrun()
        remove('invalid_spec.json')

    def test_dryrunhandler_no_dataset(self, no_datasets_specification):
        with open('invalid_spec.json', 'w') as f:
            json.dump(no_datasets_specification, f)
        with pytest.raises(ValueError):
            handler = DryRunHandler({'config_file': 'invalid_spec.json'})
            handler.valid_specification_dryrun()
        remove('invalid_spec.json')

    def test_dryrunhandler_no_dataset_size(self, invalid_no_size_dataset):
        with open('invalid_spec.json', 'w') as f:
            json.dump(invalid_no_size_dataset, f)
        with pytest.raises(ValueError):
            handler = DryRunHandler({'config_file': 'invalid_spec.json'})
            handler.valid_specification_dryrun()
        remove('invalid_spec.json')

    def test_output(self, valid_dryrun, valid_specification):
        with open('valid_spec.json', 'w') as f:
            json.dump(valid_specification, f)
        handler = BaseHandler()
        dry = DryRunHandler({'config_file': 'valid_spec.json'})
        dataframe = handler.generate_dataframe(valid_dryrun, 10)
        key = "sample"
        dataset_fields = valid_dryrun['fields']
        assert dry.print_dryrun(dataframe, key, dataset_fields) is None
        remove('valid_spec.json')

    def test_generate_dryrun(self, mocker, valid_specification):
        mock_config = mocker.patch \
                            .object(DryRunHandler,
                                    'valid_specification_dryrun')
        mock_config.return_value = valid_specification

        handler = DryRunHandler(arguments={'config_file': None})
        assert handler.generate() is None



from src.lib.mysql.mysql import MySQLConnection


class TestMysqlConnection(object):

    def test_execute(self, mocker):
        mock_conn = mocker.patch('src.lib.shell.Shell.execute')
        mock_conn.cursor.return_value = lambda x: 0

        with MySQLConnection(host='',
                             username='',
                             password='',
                             port=3306,
                             database='') as conn:
            conn.execute_query('')

        mock_conn.assert_called()
